<?php
// Own namespace
namespace Org\Mxchange\CoreFramework\Tests\Controller;

// Import framework stuff
use Org\Mxchange\CoreFramework\Controller\BaseController;
use Org\Mxchange\CoreFramework\Controller\Controller;
use Org\Mxchange\CoreFramework\Factory\Object\ObjectFactory;
use Org\Mxchange\CoreFramework\Filter\Filterable;
use Org\Mxchange\CoreFramework\Request\Requestable;
use Org\Mxchange\CoreFramework\Resolver\Command\CommandResolver;
use Org\Mxchange\CoreFramework\Response\Responseable;

/**
 * The default controller with news for e.g. home or news page
 *
 * @author		Roland Haeder <webmaster@shipsimu.org>
 * @version		0.0.0
 * @copyright	Copyright (c) 2007, 2008 Roland Haeder, 2009 - 2022 Core Developer Team
 * @license		GNU GPL 3.0 or any newer version
 * @link		http://www.shipsimu.org
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
class TestsConsoleDefaultNewsController extends BaseController implements Controller {
	/**
	 * Protected constructor
	 *
	 * @return	void
	 */
	private function __construct () {
		// Call parent constructor
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('TEST-CONSOLE-DEFAULT-NEWS-CONTROLLER: CONSTRUCTED!');
		parent::__construct(__CLASS__);

		// Init additional filter chains
		foreach (['bootstrap', 'tests', 'shutdown'] as $filterChain) {
			self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('TEST-CONSOLE-DEFAULT-NEWS-CONTROLLER: Initializing filterChain=%s ...', $filterChain));
			$this->initFilterChain($filterChain);
		}

		// Trace message
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('TEST-CONSOLE-DEFAULT-NEWS-CONTROLLER: EXIT!');
	}

	/**
	 * Creates an instance of this class
	 *
	 * @param	$resolverInstance		An instance of a command resolver class
	 * @return	$controllerInstance		A prepared instance of this class
	 */
	public static final function createTestsConsoleDefaultNewsController (CommandResolver $resolverInstance) {
		// Create the instance
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('TEST-CONSOLE-DEFAULT-NEWS-CONTROLLER: resolverInstance=%s - CALLED!', $resolverInstance->__toString()));
		$controllerInstance = new TestsConsoleDefaultNewsController();

		// Set the command resolver
		$controllerInstance->setResolverInstance($resolverInstance);

		// Add news filters to this controller
		$controllerInstance->addPreFilter(ObjectFactory::createObjectByConfiguredName('news_download_filter_class'));
		$controllerInstance->addPreFilter(ObjectFactory::createObjectByConfiguredName('news_process_filter_class'));

		// Return the prepared instance
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('TEST-CONSOLE-DEFAULT-NEWS-CONTROLLER: controllerInstance=%s - EXIT!', $controllerInstance->__toString()));
		return $controllerInstance;
	}

	/**
	 * Handles the given request and response
	 *
	 * @param	$requestInstance	An instance of a request class
	 * @param	$responseInstance	An instance of a response class
	 * @return	void
	 */
	public function handleRequest (Requestable $requestInstance, Responseable $responseInstance) {
		// Get the command instance from the resolver by sending a request instance to the resolver
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('TEST-CONSOLE-DEFAULT-NEWS-CONTROLLER: requestInstance=%s,responseInstance=%s - CALLED!', $requestInstance->__toString(), $responseInstance->__toString()));
		$commandInstance = $this->getResolverInstance()->resolveCommandByRequest($requestInstance);

		// Add more filters by the command
		$commandInstance->addExtraFilters($this, $requestInstance);

		// Run the pre filters
		$this->executePreFilters($requestInstance, $responseInstance);

		// This request was valid! :-D
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('TEST-CONSOLE-DEFAULT-NEWS-CONTROLLER: Invoking requestInstance->setIsRequestValid(TRUE) ...');
		$requestInstance->setIsRequestValid(TRUE);

		// Execute the command
		$commandInstance->execute($requestInstance, $responseInstance);

		// Run the post filters
		$this->executePostFilters($requestInstance, $responseInstance);

		// Flush the response out
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('TEST-CONSOLE-DEFAULT-NEWS-CONTROLLER: Invoking responseInstance->flushBuffer() ...');
		$responseInstance->flushBuffer();

		// Trace message
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('TEST-CONSOLE-DEFAULT-NEWS-CONTROLLER: EXIT!');
	}

	/**
	 * Add a bootstrap filter
	 *
	 * @param	$filterInstance		A Filterable class
	 * @return	void
	 */
	public function addBootstrapFilter (Filterable $filterInstance) {
		$this->addFilter('bootstrap', $filterInstance);
	}

	/**
	 * Add a tests filter
	 *
	 * @param	$filterInstance		A Filterable class
	 * @return	void
	 */
	public function addTestsFilter (Filterable $filterInstance) {
		$this->addFilter('tests', $filterInstance);
	}

	/**
	 * Executes all bootstrap filters
	 *
	 * @param	$requestInstance	A Requestable class
	 * @param	$responseInstance	A Responseable class
	 * @return	void
	 */
	public function executeBootstrapFilters (Requestable $requestInstance, Responseable $responseInstance) {
		$this->executeFilters('bootstrap', $requestInstance, $responseInstance);
	}

	/**
	 * Executes all tests filters
	 *
	 * @param	$requestInstance	A Requestable class
	 * @param	$responseInstance	A Responseable class
	 * @return	void
	 */
	public function executeTestsFilters (Requestable $requestInstance, Responseable $responseInstance) {
		$this->executeFilters('tests', $requestInstance, $responseInstance);
	}

}
